{-# LANGUAGE FlexibleContexts #-}

module Kernel where

import qualified Data.Maybe
import Control.Monad.Error.Class (MonadError, throwError)

data Sort = Kind | Type | PolymorphicTerm | Metaprop deriving (Eq, Show)

data PTerm = Free Int | Binded Int | Application PTypedTerm PTypedTerm | Lambda String PTerm Sort PTypedTerm | Pi String PTerm Sort PTerm Sort deriving (Eq, Show)
data PTypedTerm = PTypedTerm PTerm PTerm (Maybe Sort) deriving (Eq, Show)

data Statement = Statement [Maybe (String, PTerm)] PTypedTerm deriving (Show)

runMaybe :: (MonadError String m) => Maybe a -> m a
runMaybe (Just x) = return x
runMaybe Nothing = throwError "internal error: runMaybe"

pTermToSort :: (MonadError String m) => PTerm -> m Sort
pTermToSort (Free (-4)) = return Kind
pTermToSort (Free (-3)) = return Type
pTermToSort (Free (-2)) = return PolymorphicTerm
pTermToSort (Free (-1)) = return Metaprop
pTermToSort x = throwError $ "internal error: pTermToSort: " ++ show x

sortToPTerm :: Sort -> PTerm
sortToPTerm Kind            = Free (-4)
sortToPTerm Type            = Free (-3)
sortToPTerm PolymorphicTerm = Free (-2)
sortToPTerm Metaprop        = Free (-1)

sortOfSort :: Sort -> Maybe Sort
sortOfSort Kind            = Nothing
sortOfSort Type            = Just Kind
sortOfSort PolymorphicTerm = Nothing
sortOfSort Metaprop        = Just Type

typeStatement     = Statement [] (PTypedTerm (sortToPTerm Type)     (sortToPTerm Kind) Nothing)
metapropStatement = Statement [] (PTypedTerm (sortToPTerm Metaprop) (sortToPTerm Type) (Just Kind))

rule :: (MonadError String m) => Sort -> Sort -> m Sort
rule Metaprop Metaprop = return Metaprop
rule Type     Metaprop = return Metaprop
rule Kind     Metaprop = return Metaprop
rule Type     Type     = return Type
rule Kind     Type     = return PolymorphicTerm
rule a b = throwError $ "internal error: no such rule: " ++ show a ++ " and " ++ show b

merge :: (MonadError String m) => [Maybe (String, PTerm)] -> [Maybe (String, PTerm)] -> m [Maybe (String, PTerm)]

merge [] ys = return ys

merge xs [] = return xs

merge (Nothing:xs) (y:ys) = do {
  result <- merge xs ys;
  return $ y:result;
}

merge (x:xs) (Nothing:ys) = do {
  result <- merge xs ys;
  return $ x:result;
}

merge ((Just x):xs) ((Just y):ys) = if x /= y
  then throwError "internal error: merge"
  else do {
    result <- merge xs ys;
    return $ (Just x):result;
  }

replaceF      :: Int -> PTerm -> PTerm      -> PTerm
replaceFTyped :: Int -> PTerm -> PTypedTerm -> PTypedTerm

replaceF x tm (Free y) = if y == x then tm else Free y
replaceF x tm (Binded y) = Binded y
replaceF x tm (Application a b) = Application (replaceFTyped x tm a) (replaceFTyped x tm b)
replaceF x tm (Lambda y a s1 b   ) = Lambda y (replaceF x tm a) s1 (replaceFTyped x (shiftTyped 1 0 tm) b)
replaceF x tm (Pi     y a s1 b s2) = Pi     y (replaceF x tm a) s1 (replaceF      x (shiftTyped 1 0 tm) b) s2

replaceFTyped x tm (PTypedTerm a b c) = PTypedTerm (replaceF x tm a) (replaceF x tm b) c

replaceB      :: Int -> PTerm -> PTerm      -> PTerm
replaceBTyped :: Int -> PTerm -> PTypedTerm -> PTypedTerm

replaceB x tm (Free y) = Free y
replaceB x tm (Binded y) = if y == x then tm else Binded y
replaceB x tm (Application a b) = Application (replaceBTyped x tm a) (replaceBTyped x tm b)
replaceB x tm (Lambda y a s1 b   ) = Lambda y (replaceB x tm a) s1 (replaceBTyped (x + 1) (shiftTyped 1 0 tm) b)
replaceB x tm (Pi     y a s1 b s2) = Pi     y (replaceB x tm a) s1 (replaceB      (x + 1) (shiftTyped 1 0 tm) b) s2

replaceBTyped x tm (PTypedTerm a b c) = PTypedTerm (replaceB x tm a) (replaceB x tm b) c

shift      :: Int -> Int -> PTerm      -> PTerm
shiftTyped :: Int -> Int -> PTypedTerm -> PTypedTerm

shift x n (Free y) = Free y
shift x n (Binded y) = if y >= n then Binded (y + x) else Binded y
shift x n (Application a b) = Application (shiftTyped x n a) (shiftTyped x n b)
shift x n (Lambda y a s1 b   ) = Lambda y (shift x n a) s1 (shiftTyped x (n + 1) b)
shift x n (Pi     y a s1 b s2) = Pi     y (shift x n a) s1 (shift x (n + 1) b) s2

shiftTyped x (PTypedTerm a b c) = PTypedTerm (shift x a) (shift x b) c

pTermAlphaEq      :: PTerm      -> PTerm      -> Bool
pTermAlphaEqTyped :: PTypedTerm -> PTypedTerm -> Bool

pTermAlphaEq (Free a1) (Free a2) = a1 == a2
pTermAlphaEq (Binded a1) (Binded a2) = a1 == a2
pTermAlphaEq (Application a1 b1) (Application a2 b2) = pTermAlphaEqTyped a1 a2 && pTermAlphaEqTyped b1 b2
pTermAlphaEq (Lambda _ a1 _ b1) (Lambda _ a2 _ b2) = pTermAlphaEq a1 a2 && pTermAlphaEqTyped b1 b2
pTermAlphaEq (Pi _ a1 _ b1 _) (Pi _ a2 _ b2 _) = pTermAlphaEq a1 a2 && pTermAlphaEq b1 b2
pTermAlphaEq _ _ = False

pTermAlphaEqTyped (PTypedTerm a1 _ _) (PTypedTerm a2 _ _) = pTermAlphaEq a1 a2

internalAbstract :: (MonadError String m) => [Maybe (String, PTerm)] -> PTerm -> PTerm -> [Maybe (String, PTerm)] -> [PTerm] -> m ([Maybe (String, PTerm)], String, [PTerm])
internalAbstract c1 x tp1 c2 tm2 = do {
  if null c1
    then throwError "internal error: null c1"
    else return ();
  case x of {
    Free n -> if length c1 /= n + 1
      then throwError "internal error: length c1 /= n + 1";
      else return ();
    _ -> throwError "internal error: x /= Free _";
  };
  (x2, tp2) <- runMaybe $ last c1;
  if tp2 /= tp1
    then throwError "internal error: tp2 /= tp1"
    else return ();
  if length c1 < length c2
    then throwError "internal error: length c1 < length c2"
    else return ();
  result <- merge c1 c2;
  return (init result, x2, map (replaceF (length c1 - 1) (Binded 0)) tm2);
}

start :: (MonadError String m) => Int -> String -> Statement -> m Statement
start n i (Statement c (PTypedTerm tm tp _)) = if n < length c
  then throwError "internal error: start"
  else do {
    s <- pTermToSort tp;
    return $ Statement (c ++ replicate (n - length c) Nothing ++ [Just (i, tm)]) (PTypedTerm (Free n) tm (Just s));
  }

arrow :: (MonadError String m) => Statement -> Statement -> m Statement
arrow (Statement c1 (PTypedTerm x tp1 ms1)) (Statement c2 (PTypedTerm tm2 tps2 _)) = do {
  (c, x2, [tm2b]) <- internalAbstract c1 x tp1 c2 [tm2];
  s1 <- runMaybe ms1;
  s2 <- pTermToSort tps2;
  s3 <- rule s1 s2;
  return $ Statement c (PTypedTerm (Pi x2 tp1 s1 tm2b s2) (sortToPTerm s3) (sortOfSort s3));
}

application :: (MonadError String m) => Statement -> Statement -> m Statement
application (Statement c1 (PTypedTerm tm1 piType s1)) (Statement c2 (PTypedTerm tm2 tp2 s2)) = do {
  c <- merge c1 c2;
  case piType of {
    Pi _ shouldBeTp2 _ tpRes sRes -> do {
      if pTermAlphaEq shouldBeTp2 tp2
        then return ()
        else throwError ("internal error: application: type mismatch. " ++ show (Statement c1 (PTypedTerm tm1 piType s1)) ++ " and " ++ show (Statement c2 (PTypedTerm tm2 tp2 s2)));
      return $ Statement c (PTypedTerm (Application (PTypedTerm tm1 piType s1) (PTypedTerm tm2 tp2 s2)) (replaceB 0 tm2 tpRes) (Just sRes));
    };
    _ -> throwError "internal error: application: not a pi";
  };
}

abstraction :: (MonadError String m) => Statement -> Statement -> m Statement
abstraction (Statement c1 (PTypedTerm x a mas)) (Statement c2 (PTypedTerm b bt mbs)) = do {
  (c, x2, [bq, btq]) <- internalAbstract c1 x a c2 [b, bt];
  as <- runMaybe mas;
  bs <- runMaybe mbs;
  s3 <- rule as bs;
  return $ Statement c (PTypedTerm (Lambda x2 a as (PTypedTerm bq btq mbs)) (Pi x2 a as btq bs) (Just s3));
}

sortOf :: (MonadError String m) => Statement -> m Sort
sortOf (Statement _ (PTypedTerm _ tp _)) = pTermToSort tp

typeOf :: (MonadError String m) => Statement -> m Statement
typeOf (Statement c (PTypedTerm _ tp (Just s))) = return $ Statement c (PTypedTerm tp (sortToPTerm s) (sortOfSort s))
typeOf _ = throwError "this PTS-term has no type"

statementEq :: (MonadError String m) => Statement -> Statement -> m Bool
statementEq (Statement c1 (PTypedTerm tm1 _ _)) (Statement c2 (PTypedTerm tm2 _ _)) = do {
  merge c1 c2;
  return $ pTermAlphaEq tm1 tm2;
}

statementCase :: Statement -> (String -> Statement -> a) -> (Statement -> Statement -> a) -> (String -> Statement -> Statement -> a) -> (String -> Statement -> Statement -> a) -> a
statementCase (Statement c (PTypedTerm (Free x         ) a b)) f _ _ _ = if x < 0
  then f "$sort" (Statement [] (PTypedTerm (Free x) a b))
  else let Just result = c !! x in f (fst result) (Statement (take (x + 1) c) (PTypedTerm (Free x) a b))
statementCase (Statement c (PTypedTerm (Binded _       ) _ _)) _ _ _ _ = error "free binded o_O"
statementCase (Statement c (PTypedTerm (Application a b) _ _)) _ f _ _ = f (Statement c a) (Statement c b)
statementCase (Statement c (PTypedTerm (Lambda x a as b) _ _)) _ _ f _ = f x (Statement c (PTypedTerm a (sortToPTerm as) (sortOfSort as))) (Statement (c ++ [Just (x, a)]) $ replaceBTyped 0 (Free $ length c) b)
statementCase (Statement c (PTypedTerm (Pi x a as b bs ) _ _)) _ _ _ f = f x (Statement c (PTypedTerm a (sortToPTerm as) (sortOfSort as))) (Statement (c ++ [Just (x, a)]) (PTypedTerm (replaceB 0 (Free $ length c) b) (sortToPTerm bs) (sortOfSort bs)))
