{-# LANGUAGE FlexibleContexts #-}

import qualified Data.Map

import qualified Control.Monad
import Control.Monad.Error.Class (MonadError, throwError)

import Text.Parsec
import qualified Text.Parsec.Error
import qualified Text.Parsec.Token

import Kernel

data ExprTable s st m a = ExprTable (Data.Map.Map Int ([(Int -> ParsecT s st m a) -> ParsecT s st m a], [(Int -> ParsecT s st m a) -> a -> ParsecT s st m a]))

instance Show (ExprTable s st m a) where
  show _ = "(ExprTable)"

exprEmpty :: ExprTable s st m a
exprEmpty = ExprTable Data.Map.empty

exprAddNormal :: Int -> ((Int -> ParsecT s st m a) -> ParsecT s st m a) -> ExprTable s st m a -> ExprTable s st m a
exprAddNormal priority prod (ExprTable m) = case Data.Map.lookup priority m of {
  Nothing -> ExprTable $ Data.Map.insert priority ([prod], []) m;
  Just (n, r) -> ExprTable $ Data.Map.insert priority (prod:n, r) m;
}

exprAddRecursive :: Int -> ((Int -> ParsecT s st m a) -> a -> ParsecT s st m a) -> ExprTable s st m a -> ExprTable s st m a
exprAddRecursive priority prod (ExprTable m) = case Data.Map.lookup priority m of {
  Nothing -> ExprTable $ Data.Map.insert priority ([], [prod]) m;
  Just (n, r) -> ExprTable $ Data.Map.insert priority (n, prod:r) m;
}

exprParsePriority :: ExprTable s st m a -> Int -> ParsecT s st m a
exprParsePriority (ExprTable m) priority = case Data.Map.minViewWithKey $ Data.Map.filterWithKey (\k _ -> k >= priority) m of {
  Nothing -> parserZero;
  Just ((k, (normalProductions, recursiveProductions)), _) -> do {
    first <- (foldl (<|>) parserZero $ map (\f -> try $ f (exprParsePriority $ ExprTable m)) normalProductions) <|> exprParsePriority (ExprTable m) (k + 1);
    let recursives x = do {
      result <- foldl (<|>) parserZero $ map (\f -> try $ f (exprParsePriority $ ExprTable m) x) recursiveProductions;
      recursives result;
    } <|> return x in recursives first;
  };
}

exprParse :: ExprTable s st m a -> ParsecT s st m a
exprParse table = exprParsePriority table 0

data SyntacticPTerm = SConst Statement | SVar String | SOfType SyntacticPTerm SyntacticPTerm | SApplication SyntacticPTerm SyntacticPTerm Statement | SPi String SyntacticPTerm SyntacticPTerm Statement deriving (Show)

type InnerLowLevelMonad = Either String
type OuterLowLevelMonad = Either String

data InnerState = InnerState {
  stN :: Int,
  stC :: [(String, Statement)],
  stJ :: Maybe Statement,
  stSyntax :: ExprTable String InnerState InnerLowLevelMonad SyntacticPTerm
} deriving (Show)

type OuterState = InnerState

type InnerParserMonad = ParsecT String InnerState InnerLowLevelMonad
type OuterParserMonad = ParsecT String OuterState OuterLowLevelMonad

innerLangDef :: Text.Parsec.Token.GenLanguageDef String InnerState InnerLowLevelMonad
innerLangDef = Text.Parsec.Token.LanguageDef {
  Text.Parsec.Token.commentStart    = "",
  Text.Parsec.Token.commentEnd      = "",
  Text.Parsec.Token.commentLine     = "",
  Text.Parsec.Token.nestedComments  = False,
  Text.Parsec.Token.identStart      = letter <|> oneOf "_'",
  Text.Parsec.Token.identLetter     = alphaNum <|> oneOf "_'",
  Text.Parsec.Token.opStart         = parserZero,
  Text.Parsec.Token.opLetter        = parserZero,
  Text.Parsec.Token.reservedOpNames = ["=>", "::", "==>"],
  Text.Parsec.Token.reservedNames   = ["_"],
  Text.Parsec.Token.caseSensitive   = True
}

outerLangDef :: Text.Parsec.Token.GenLanguageDef String OuterState OuterLowLevelMonad
outerLangDef = Text.Parsec.Token.LanguageDef {
  Text.Parsec.Token.commentStart    = "(*",
  Text.Parsec.Token.commentEnd      = "*)",
  Text.Parsec.Token.commentLine     = "",
  Text.Parsec.Token.nestedComments  = False,
  Text.Parsec.Token.identStart      = letter <|> oneOf "_'",
  Text.Parsec.Token.identLetter     = alphaNum <|> oneOf "_'",
  Text.Parsec.Token.opStart         = parserZero,
  Text.Parsec.Token.opLetter        = parserZero,
  Text.Parsec.Token.reservedOpNames = ["::", ":", "{", "}"],
  Text.Parsec.Token.reservedNames   = ["_", "consts", "judgment", "typedecl", "prefix", "infix", "infixl", "infixr", "binder", "axioms", "theorem", "lemma", "assumes", "shows", "by", "proof", "show", "have", "qed", "assume"],
  Text.Parsec.Token.caseSensitive   = True
}

innerLexer :: Text.Parsec.Token.GenTokenParser String InnerState InnerLowLevelMonad
innerLexer = Text.Parsec.Token.makeTokenParser innerLangDef

outerLexer :: Text.Parsec.Token.GenTokenParser String OuterState OuterLowLevelMonad
outerLexer = Text.Parsec.Token.makeTokenParser outerLangDef

iIdentifier    = Text.Parsec.Token.identifier    innerLexer
iWhiteSpace    = Text.Parsec.Token.whiteSpace    innerLexer
iParens        = Text.Parsec.Token.parens        innerLexer
iReservedOp    = Text.Parsec.Token.reservedOp    innerLexer

oIdentifier    = Text.Parsec.Token.identifier    outerLexer
oWhiteSpace    = Text.Parsec.Token.whiteSpace    outerLexer
oReservedOp    = Text.Parsec.Token.reservedOp    outerLexer
oReserved      = Text.Parsec.Token.reserved      outerLexer
oDecimal       = Text.Parsec.Token.decimal       outerLexer

oStringLiteral = Text.Parsec.Token.stringLiteral outerLexer

iTypeAtom :: InnerParserMonad SyntacticPTerm
iType     :: InnerParserMonad SyntacticPTerm

iTypeAtom = do {
  st <- getState;
  i <- iIdentifier;
  case lookup i $ reverse $ stC st of {
    Nothing -> if i !! 0 == '\''
      then do {
        result <- start (stN st) i typeStatement;
        setState $ st { stN = stN st + 1, stC = stC st ++ [(i, result)] };
        return $ SConst result;
      }
      else throwError (i ++ " undeclared");
    Just result -> do {
      resultType <- typeOf result;
      eq <- statementEq resultType typeStatement;
      if eq
        then return (SConst result)
        else throwError (i ++ " is not a type");
    };
  };
} <|> iParens iType

iType = do {
  arg <- iTypeAtom;
  do {
    iReservedOp "=>";
    res <- iType;
    return $ SPi "_" arg res typeStatement;
  } <|> return arg;
}

iTypeParser :: InnerParserMonad (SyntacticPTerm, InnerState)
iTypeParser = do {
  iWhiteSpace;
  result <- iType;
  eof;
  st <- getState;
  return (result, st);
}

myRunParser :: (MonadError String m) => ParsecT String st (Either String) ret -> st -> String -> String -> m ret
myRunParser parser st document lang = case runParserT parser st "-" document of {
  Left nonParserErr -> throwError nonParserErr;
  Right result -> case result of {
    Left parserErr -> throwError $ lang ++ ": " ++ Text.Parsec.Error.showErrorMessages "or" "???" "expected" "found" "nothing" (Text.Parsec.Error.errorMessages parserErr);
    Right result -> return result;
  };
}

oType :: OuterParserMonad SyntacticPTerm
oType = do {
  st <- getState;
  str <- oIdentifier <|> oStringLiteral;
  (result, st2) <- myRunParser iTypeParser st str ("type \"" ++ str ++ "\"");
  setState st2;
  return result;
}

data UPTerm = UConst Statement | UUnknown Int | UVar String | UApplication UPTerm UPTerm | UPi String UPTerm UPTerm deriving (Show)

occursCheck :: (MonadError String m) => Data.Map.Map Int UPTerm -> Int -> UPTerm -> m ()
occursCheck _ _      (UConst _        ) = return ()
occursCheck m needle (UUnknown n      ) = if needle == n
  then throwError "occurs check failed"
  else case Data.Map.lookup n m of {
    Nothing -> return ();
    Just a -> occursCheck m needle a;
  }
occursCheck _ needle (UVar _          ) = return ()
occursCheck m needle (UApplication a b) = do {
  occursCheck m needle a;
  occursCheck m needle b;
}
occursCheck m needle (UPi x a b       ) = do {
  occursCheck m needle a;
  occursCheck m needle b;
}

unifyVar :: (MonadError String m) => Data.Map.Map Int UPTerm -> Int    -> UPTerm -> m (Data.Map.Map Int UPTerm)
unify    :: (MonadError String m) => Data.Map.Map Int UPTerm -> UPTerm -> UPTerm -> m (Data.Map.Map Int UPTerm)

unifyVar m an b = case Data.Map.lookup an m of {
  Nothing -> case b of {
    UUnknown bn -> if an == bn
      then return m
      else return (Data.Map.insert an b m);
    _ -> do {
      occursCheck m an b;
      return $ Data.Map.insert an b m;
    };
  };
  Just a -> unify m a b;
};

unify m (UUnknown an) b = unifyVar m an b
unify m a (UUnknown bn) = unifyVar m bn a

unify m (UConst a) (UConst b) = do {
  eq <- statementEq a b;
  if eq
    then return m
    else throwError ("uf: UConst: " ++ show a ++ " and " ++ show b);
}

unify m (UVar a) (UVar b) = if a == b
  then return m
  else throwError "uf: UVar"

unify m (UApplication a1 a2) (UApplication b1 b2) = do {
  m2 <- unify m a1 b1;
  unify m2 a2 b2;
}

unify m (UPi _ a1 a2) (UPi _ b1 b2) = do {
  m2 <- unify m a1 b1;
  unify m2 a2 b2;
}

unify _ a b = throwError $ "uf: other: " ++ show a ++ " and " ++ show b

pTermToUTerm :: Data.Map.Map String Int -> Statement -> UPTerm
pTermToUTerm m tp = statementCase tp
  (\x tp2 -> case Data.Map.lookup x m of {
    Nothing -> UConst tp2;
    Just n -> UUnknown n;
  })
  (\a b -> UApplication (pTermToUTerm m a) (pTermToUTerm m b))
  (\_ _ _ -> UConst tp)
  (\x a b -> UPi "_" (pTermToUTerm m a) (pTermToUTerm m b))

data UCtx = UCtx {
  uUs :: [UPTerm],
  uVs :: Data.Map.Map String UPTerm,
  uEqs :: Data.Map.Map Int UPTerm
} deriving (Show)

uCtxEmpty :: UCtx
uCtxEmpty = UCtx { uUs = [], uVs = Data.Map.empty, uEqs = Data.Map.empty }

constToUTerm :: (MonadError String m) => UCtx -> Data.Map.Map String Int -> UPTerm -> Statement -> m (UCtx, UPTerm, UPTerm)
constToUTerm ctx m tm tp = statementCase tp
  (\_ _ -> return (ctx, tm, pTermToUTerm m tp))
  (\_ _ -> return (ctx, tm, pTermToUTerm m tp))
  (\_ _ _ -> return (ctx, tm, pTermToUTerm m tp))
  (\x xt tm2 -> do {
    xts <- sortOf xt;
    tm2s <- sortOf tm2;
    if (xts, tm2s) == (Kind, Type) || (xts, tm2s) == (Kind, Metaprop) || (xts, tm2s) == (Type, Metaprop)
      then do {
        tp2 <- return $ length $ uUs ctx;
        ctx2 <- return $ ctx { uUs = uUs ctx ++ [UConst xt] };
        constToUTerm ctx2 (Data.Map.insert x tp2 m) (UApplication tm (UUnknown tp2)) tm2;
      }
      else return (ctx, tm, pTermToUTerm m tp);
  })

typeCheck :: (MonadError String m) => UCtx -> SyntacticPTerm -> m (UCtx, UPTerm, UPTerm)
typeCheck ctx (SConst tm) = do {
  tp <- typeOf tm;
  let m = Data.Map.empty in statementCase tp
    (\_ _ -> return (ctx, pTermToUTerm m tm, pTermToUTerm m tp))
    (\_ _ -> return (ctx, pTermToUTerm m tm, pTermToUTerm m tp))
    (\_ _ _ -> return (ctx, pTermToUTerm m tm, pTermToUTerm m tp))
    (\x xt tm2 -> do {
      xts <- sortOf xt;
      tm2s <- sortOf tm2;
      if (xts, tm2s) == (Kind, Type) || (xts, tm2s) == (Kind, Metaprop) || (xts, tm2s) == (Type, Metaprop)
        then constToUTerm ctx m (UConst tm) tp
        else return (ctx, pTermToUTerm m tm, pTermToUTerm m tp);
    });
}
typeCheck ctx (SVar i) = case Data.Map.lookup i $ uVs ctx of {
  Nothing -> do {
    tp <- return $ length $ uUs ctx;
    return (ctx { uUs = uUs ctx ++ [UConst typeStatement], uVs = Data.Map.insert i (UUnknown tp) $ uVs ctx }, UVar i, UUnknown tp);
  };
  Just tp -> return (ctx, UVar i, tp);
}
typeCheck ctx (SOfType a b) = do {
  (ctx2, aa, at) <- typeCheck ctx a;
  (ctx3, bb, _ ) <- typeCheck ctx2 b;
  m <- unify (uEqs ctx3) at bb;
  return (ctx3 { uEqs = m }, aa, bb);
}
typeCheck ctx (SApplication a b s) = do {
  (ctx2, aa, at) <- typeCheck ctx a;
  (ctx3, bb, bt) <- typeCheck ctx2 b;
  tp <- return $ length $ uUs ctx3;
  ctx4 <- return $ ctx3 { uUs = uUs ctx3 ++ [UConst s] };
  m <- unify (uEqs ctx4) at (UPi "_" bt (UUnknown tp));
  return (ctx4 { uEqs = m }, UApplication aa bb, UUnknown tp);
}
typeCheck ctx (SPi x a b s) = do {
  (ctx2, aa, _) <- typeCheck ctx a;
  (ctx3, bb, _) <- typeCheck ctx2 b;
  return (ctx3, UPi "_" aa bb, UConst s)
}

iAtomMetaprop :: InnerParserMonad SyntacticPTerm
iAtomMetaprop = do {
  st <- getState;
  j <- runMaybe (stJ st);
  a <- exprParse $ stSyntax st;
  return $ SApplication (SConst j) a metapropStatement;
} <|> iParens iMetaprop

iMetaprop :: InnerParserMonad SyntacticPTerm
iMetaprop = do {
  a <- iAtomMetaprop;
  do {
    oReservedOp "==>";
    b <- iMetaprop;
    return $ SPi "_" a b metapropStatement;
  } <|> return a;
}

iMetapropParser :: InnerParserMonad (SyntacticPTerm, InnerState)
iMetapropParser = do {
  iWhiteSpace;
  result <- iMetaprop;
  eof;
  st <- getState;
  return (result, st);
}

oMetaprop :: OuterParserMonad SyntacticPTerm
oMetaprop = do {
  st <- getState;
  str <- oIdentifier <|> oStringLiteral;
  (result, st2) <- myRunParser iMetapropParser st str ("metaprop \"" ++ str ++ "\"");
  setState st2;
  return result;
}

uPTermToStatement :: (MonadError String m) => Int -> Data.Map.Map Int Statement -> Data.Map.Map String Statement -> UPTerm -> m Statement
uPTermToStatement n us vs (UConst a) = return a
uPTermToStatement n us vs (UUnknown a) = runMaybe (Data.Map.lookup a us)
uPTermToStatement n us vs (UVar a) = runMaybe (Data.Map.lookup a vs)
uPTermToStatement n us vs (UApplication a b) = do {
  aa <- uPTermToStatement n us vs a;
  bb <- uPTermToStatement n us vs b;
  application aa bb;
}
uPTermToStatement n us vs (UPi x a b) = do {
  aa <- uPTermToStatement n us vs a;
  xx <- start n x aa;
  bb <- uPTermToStatement (n + 1) us (Data.Map.insert x xx vs) b;
  arrow xx bb;
}

oStartUs :: Data.Map.Map Int Statement -> Data.Map.Map Int UPTerm -> [(Int, UPTerm)] -> OuterParserMonad ([(Int, UPTerm)], Data.Map.Map Int Statement)
oStartUs m _ [] = return ([], m)
oStartUs m eqs ((n, tp):list) = do {
  st <- getState;
  case Data.Map.lookup n eqs of {
    Nothing -> case uPTermToStatement (stN st) m Data.Map.empty tp of {
      Left "internal error: runMaybe" -> do {
        (left, result) <- oStartUs m eqs list;
        return ((n, tp):left, result);
      };
      Left err -> throwError err;
      Right tp2 -> do {
        x <- start (stN st) ("?" ++ show n) tp2;
        setState $ st { stN = stN st + 1, stC = stC st ++ [("?" ++ show n, x)] };
        oStartUs (Data.Map.insert n x m) eqs list;
      };
    };
    Just eq -> case uPTermToStatement (stN st) m Data.Map.empty eq of {
      Left "internal error: runMaybe" -> do {
        (left, result) <- oStartUs m eqs list;
        return ((n, tp):left, result);
      };
      Left err -> throwError err;
      Right eq2 -> oStartUs (Data.Map.insert n eq2 m) eqs list;
    };
  };
}

oStartUsRepeated :: Data.Map.Map Int Statement -> Data.Map.Map Int UPTerm -> [(Int, UPTerm)] -> OuterParserMonad (Data.Map.Map Int Statement)
oStartUsRepeated m eqs list = do {
  (left, result) <- oStartUs m eqs list;
  case left of {
    [] -> return result;
    _ -> oStartUsRepeated result eqs left;
  };
}

oStartVs :: Data.Map.Map Int Statement -> Data.Map.Map String UPTerm -> OuterParserMonad (Data.Map.Map String Statement)
oStartVs m vs = case Data.Map.minViewWithKey vs of {
  Nothing -> return Data.Map.empty;
  Just ((x, tp), left) -> do {
    st <- getState;
    tp2 <- uPTermToStatement (stN st) m Data.Map.empty tp;
    xx <- start (stN st) x tp2;
    setState $ st { stN = stN st + 1, stC = stC st ++ [(x, xx)] };
    result <- oStartVs m left;
    return $ Data.Map.insert x xx result;
  };
}

oSolve :: UCtx -> UPTerm -> OuterParserMonad Statement
oSolve ctx tm = do {
  m <- oStartUsRepeated Data.Map.empty (uEqs ctx) (zip [0..] (uUs ctx));
  m2 <- oStartVs m (uVs ctx);
  st <- getState;
  uPTermToStatement (stN st) m m2 tm;
}

oConst :: OuterParserMonad ()
oConst = do {
  st1 <- getState;
  i <- oIdentifier;
  case lookup i $ reverse $ stC st1 of {
    Nothing -> return ();
    Just _ -> throwError $ i ++ ": redeclaration";
  };
  oReservedOp "::";
  syntacticType <- oType;
  (ctx, uType, _) <- typeCheck uCtxEmpty syntacticType;
  tp <- oSolve ctx uType;
  st2 <- getState;
  polymorphicType <- Control.Monad.foldM (flip arrow) tp (reverse $ drop (length $ stC st1) $ map snd $ stC st2);
  cnst <- start (stN st1) i polymorphicType;

  newSyntax <- iParens (do {
    s <- oStringLiteral;
    return $ exprAddNormal 1000 (\f -> do {
      iReservedOp s;
      return $ SConst cnst;
    }) $ stSyntax st1;
  } <|> do {
    oReserved "prefix";
    s <- oStringLiteral;
    priority <- oDecimal;
    return $ exprAddNormal (fromInteger priority) (\f -> do {
      iReservedOp s;
      arg <- f (fromInteger priority);
      return $ SApplication (SConst cnst) arg typeStatement;
    }) $ stSyntax st1;
  } <|> do {
    oReserved "infix";
    s <- oStringLiteral;
    priority <- oDecimal;
    return $ exprAddNormal (fromInteger priority) (\f -> do {
      a <- f (fromInteger priority + 1);
      iReservedOp s;
      b <- f (fromInteger priority + 1);
      return $ SApplication (SApplication (SConst cnst) a typeStatement) b typeStatement;
    }) $ stSyntax st1;
  } <|> do {
    oReserved "infixl";
    s <- oStringLiteral;
    priority <- oDecimal;
    return $ exprAddRecursive (fromInteger priority) (\f a -> do {
      iReservedOp s;
      b <- f (fromInteger priority + 1);
      return $ SApplication (SApplication (SConst cnst) a typeStatement) b typeStatement;
    }) $ stSyntax st1;
  } <|> do {
    oReserved "infixr";
    s <- oStringLiteral;
    priority <- oDecimal;
    return $ exprAddNormal (fromInteger priority) (\f -> do {
      a <- f (fromInteger priority + 1);
      iReservedOp s;
      b <- f (fromInteger priority);
      return $ SApplication (SApplication (SConst cnst) a typeStatement) b typeStatement;
    }) $ stSyntax st1;
  } <|> do {
    oReserved "binder";
    undefined;
  }) <|> return (stSyntax st1);

  setState $ st1 { stN = stN st1 + 1, stC = stC st1 ++ [(i, cnst)], stSyntax = newSyntax };
  return ();
}

iVar :: InnerParserMonad SyntacticPTerm
iVar = do {
  st <- getState;
  i <- iIdentifier;
  case lookup i $ reverse $ stC st of {
    Nothing -> return $ SVar i;
    Just result -> return $ SConst result;
  };
}

oSolveMetaprop :: OuterParserMonad Statement
oSolveMetaprop = do {
  a <- oMetaprop;
  (ctx, a2, _) <- typeCheck uCtxEmpty a;
  oSolve ctx a2;
}

oAxiom :: OuterParserMonad ()
oAxiom = do {
  st1 <- getState;
  i <- oIdentifier;
  case lookup i $ reverse $ stC st1 of {
    Nothing -> return ();
    Just _ -> throwError $ i ++ ": redeclaration";
  };
  oReservedOp ":";
  ptm <- oSolveMetaprop;
  st2 <- getState;
  polymorphicPTm <- Control.Monad.foldM (flip arrow) ptm (reverse $ drop (length $ stC st1) $ map snd $ stC st2);
  cnst <- start (stN st1) i polymorphicPTm;
  setState $ st1 { stN = stN st1 + 1, stC = stC st1 ++ [(i, cnst)] };
  return ();
}

iAtomProofTerm :: InnerParserMonad SyntacticPTerm
iAtomProofTerm = do {
  st <- getState;
  i <- iIdentifier;
  result <- runMaybe $ lookup i $ reverse $ stC st;
  return $ SConst result;
} <|> iParens iProofTerm;

iProofTerm :: InnerParserMonad SyntacticPTerm
iProofTerm = do {
  a <- iAtomProofTerm;
  b <- many iAtomProofTerm;
  return $ foldl (\x y -> SApplication x y metapropStatement) a b;
}

iProofTermParser :: InnerParserMonad SyntacticPTerm
iProofTermParser = do {
  iWhiteSpace;
  result <- iProofTerm;
  eof;
  return result;
}

oProofTerm :: OuterParserMonad SyntacticPTerm
oProofTerm = do {
  st <- getState;
  str <- oIdentifier <|> oStringLiteral;
  result <- myRunParser iProofTermParser st str ("proof term \"" ++ str ++ "\"");
  return result;
}

oNoNewMetaprop :: OuterParserMonad Statement
oNoNewMetaprop = do {
  stBegin <- getState;
  a <- oSolveMetaprop;
  stEnd <- getState;
  if length (stC stBegin) /= length (stC stEnd)
    then throwError "new vars in metaprop"
    else return ();
  return a;
}

oStep :: OuterParserMonad ()
oStep = do {
  oReserved "have";
  i <- try (do {
    i <- oIdentifier;
    oReservedOp ":";
    return i;
  }) <|> return "this";
  a <- oNoNewMetaprop;
  a2 <- oProof a;
  st <- getState;
  setState $ st { stC = stC st ++ [(i, a2), ("this", a2)] };
  return ();
} <|> do {
  stBegin <- getState;
  oReservedOp "{";
  llist <- many1 $ do {
    oStep;
    return [];
  } <|> do {
    oReserved "assume";
    i <- try (do {
      i <- oIdentifier;
      oReservedOp ":";
      return i;
    }) <|> return "this";
    a <- oNoNewMetaprop;
    st <- getState;
    a2 <- start (stN st) i a;
    setState $ st { stN = stN st + 1, stC = stC st ++ [(i, a2), ("this", a2), ("assm", a2)] };
    return [a2];
  };
  list <- Control.Monad.foldM (\x y -> return $ x ++ y) [] llist;
  oReservedOp "}";
  stEnd <- getState;
  a <- runMaybe $ lookup "this" $ reverse $ stC stEnd;
  a2 <- Control.Monad.foldM (flip abstraction) a (reverse $ list);
  setState $ stBegin { stC = stC stBegin ++ [("this", a2)] };
  return ();
}

oProof :: Statement -> OuterParserMonad Statement
oProof prop = do {
  st1 <- getState;
  oReserved "by";
  proof <- oProofTerm;
  (ctx, proof2, _) <- typeCheck uCtxEmpty (SOfType proof (SConst prop));
  proof3 <- oSolve ctx proof2;
  st2 <- getState;
  if length (stC st1) /= length (stC st2)
    then throwError "свободные переменные в by, нельзя, т. к. могут быть пустые типы"
    else return ();
  return proof3;
} <|> do {
  stBegin <- getState;
  oReserved "proof";
  many oStep;
  oReserved "show";
  result <- oProof prop;
  oReserved "qed";
  setState stBegin;
  return result;
}

oParser :: OuterParserMonad ()
oParser = do {
  oWhiteSpace;
  many (do {
    oReserved "consts";
    many1 oConst;
    return ();
  } <|> do {
    oReserved "judgment";
    oConst;
    st <- getState;
    setState $ st { stJ = Just $ snd $ last $ stC st };
    return ();
  } <|> do {
    st <- getState;
    oReserved "typedecl";
    i <- oIdentifier;
    tp <- start (stN st) i typeStatement;
    setState $ st { stN = stN st + 1, stC = stC st ++ [(i, tp)] };
    return ();
  } <|> do {
    oReserved "axioms";
    many1 oAxiom;
    return ();
  } <|> do {
    stBegin <- getState;
    oReserved "theorem" <|> oReserved "lemma";
    i <- oIdentifier;
    oReservedOp ":";
    (a, s) <- do {
      assms <- many (do {
        oReserved "assumes";
        i <- oIdentifier;
        oReservedOp ":";
        a <- oMetaprop;
        return (i, a);
      });
      oReserved "shows";
      a <- oMetaprop;
      return (assms, a);
    } <|> do {
      a <- oMetaprop;
      return ([], a);
    };
    (a2, ctx) <- Control.Monad.foldM (\(list, ctx) (i, assm) -> do {
      (ctx2, result, _) <- typeCheck ctx assm;
      return $ (list ++ [(i, result)], ctx2);
    }) ([], uCtxEmpty) a;
    (ctx2, s2, _) <- typeCheck ctx s;
    (a3, s3) <- do {
      m <- oStartUsRepeated Data.Map.empty (uEqs ctx2) (zip [0..] (uUs ctx2));
      m2 <- oStartVs m (uVs ctx2);
      st <- getState;
      a3 <- mapM (\(i, a) -> do {
        result <- uPTermToStatement (stN st) m m2 a;
        return (i, result);
      }) a2;
      s3 <- uPTermToStatement (stN st) m m2 s2;
      return (a3, s3);
    };
    mapM (\(i, a) -> do {
      st <- getState;
      x <- start (stN st) i a;
      setState $ st { stN = stN st + 1, stC = stC st ++ [(i, x)] };
      return ();
    }) a3;
    proof <- oProof s3;
    stEnd <- getState;
    theorem <- Control.Monad.foldM (flip abstraction) proof (reverse $ drop (length $ stC stBegin) $ map snd $ stC stEnd);
    setState $ stBegin { stC = stC stBegin ++ [(i, theorem)] };
    return ();
  });
  eof;
  return ();
}

main = do {
  document <- getContents;
  case myRunParser oParser (InnerState {
    stN = 0,
    stC = [("type", typeStatement), ("metaprop", metapropStatement)],
    stJ = Nothing,
    stSyntax = exprAddNormal 1000 (\f -> iVar) $
      exprAddNormal 1000 (\f -> iParens (f 0)) $
      exprAddNormal 3 (\f -> do {
        a <- f 4;
        iReservedOp "::";
        t <- iType;
        return $ SOfType a t;
      }) $
      exprAddRecursive 999 (\f func -> do {
        arg <- f 1000;
        return $ SApplication func arg typeStatement;
      }) exprEmpty
  }) document "outer language" of {
    Left err -> error err;
    Right () -> return ();
  };
}
