{-# LANGUAGE FlexibleContexts #-}

import qualified Data.Map

import qualified Control.Monad
import Control.Monad.Error.Class (MonadError, throwError, catchError)

import Text.Parsec
import qualified Text.Parsec.Error
import qualified Text.Parsec.Token

import Kernel

data ExprTable s st m a = ExprTable (Data.Map.Map Int ([(Int -> ParsecT s st m a) -> ParsecT s st m a], [(Int -> ParsecT s st m a) -> a -> ParsecT s st m a]))

instance Show (ExprTable s st m a) where
  show _ = "(ExprTable)"

exprEmpty :: ExprTable s st m a
exprEmpty = ExprTable Data.Map.empty

exprAddNormal :: Int -> ((Int -> ParsecT s st m a) -> ParsecT s st m a) -> ExprTable s st m a -> ExprTable s st m a
exprAddNormal priority prod (ExprTable m) = case Data.Map.lookup priority m of {
  Nothing -> ExprTable $ Data.Map.insert priority ([prod], []) m;
  Just (n, r) -> ExprTable $ Data.Map.insert priority (prod:n, r) m;
}

exprAddRecursive :: Int -> ((Int -> ParsecT s st m a) -> a -> ParsecT s st m a) -> ExprTable s st m a -> ExprTable s st m a
exprAddRecursive priority prod (ExprTable m) = case Data.Map.lookup priority m of {
  Nothing -> ExprTable $ Data.Map.insert priority ([], [prod]) m;
  Just (n, r) -> ExprTable $ Data.Map.insert priority (n, prod:r) m;
}

exprParsePriority :: ExprTable s st m a -> Int -> ParsecT s st m a
exprParsePriority (ExprTable m) priority = case Data.Map.minViewWithKey $ Data.Map.filterWithKey (\k _ -> k >= priority) m of {
  Nothing -> parserZero;
  Just ((k, (normalProductions, recursiveProductions)), _) -> do {
    first <- (foldl (<|>) parserZero $ map (\f -> try $ f (exprParsePriority $ ExprTable m)) normalProductions) <|> exprParsePriority (ExprTable m) (k + 1);
    let recursives x = do {
      result <- foldl (<|>) parserZero $ map (\f -> try $ f (exprParsePriority $ ExprTable m) x) recursiveProductions;
      recursives result;
    } <|> return x in recursives first;
  };
}

exprParse :: ExprTable s st m a -> ParsecT s st m a
exprParse table = exprParsePriority table 0

type InnerLowLevelMonad = Either String
type OuterLowLevelMonad = Either String

data InnerState = InnerState {
  stN :: Int,
  stC :: [(String, Statement)],
  stSyntax :: ExprTable String InnerState InnerLowLevelMonad Statement
} deriving (Show)

type OuterState = InnerState

type InnerParserMonad = ParsecT String InnerState InnerLowLevelMonad
type OuterParserMonad = ParsecT String OuterState OuterLowLevelMonad

innerLangDef :: Text.Parsec.Token.GenLanguageDef String InnerState InnerLowLevelMonad
innerLangDef = Text.Parsec.Token.LanguageDef {
  Text.Parsec.Token.commentStart    = "",
  Text.Parsec.Token.commentEnd      = "",
  Text.Parsec.Token.commentLine     = "",
  Text.Parsec.Token.nestedComments  = False,
  Text.Parsec.Token.identStart      = letter <|> oneOf "_",
  Text.Parsec.Token.identLetter     = alphaNum <|> oneOf "_",
  Text.Parsec.Token.opStart         = parserZero,
  Text.Parsec.Token.opLetter        = parserZero,
  Text.Parsec.Token.reservedOpNames = ["::", ".", "%", "!!", "==>"],
  Text.Parsec.Token.reservedNames   = ["_"],
  Text.Parsec.Token.caseSensitive   = True
}

outerLangDef :: Text.Parsec.Token.GenLanguageDef String OuterState OuterLowLevelMonad
outerLangDef = Text.Parsec.Token.LanguageDef {
  Text.Parsec.Token.commentStart    = "(*",
  Text.Parsec.Token.commentEnd      = "*)",
  Text.Parsec.Token.commentLine     = "",
  Text.Parsec.Token.nestedComments  = False,
  Text.Parsec.Token.identStart      = letter <|> oneOf "_",
  Text.Parsec.Token.identLetter     = alphaNum <|> oneOf "_",
  Text.Parsec.Token.opStart         = parserZero,
  Text.Parsec.Token.opLetter        = parserZero,
  Text.Parsec.Token.reservedOpNames = ["::", "=", "{", "}", "(", ")", ","],
  Text.Parsec.Token.reservedNames   = ["_", "prefix", "infix", "infixl", "infixr", "binder", "fix", "abbr", "sorts", "axioms", "rules"],
  Text.Parsec.Token.caseSensitive   = True
}

innerLexer :: Text.Parsec.Token.GenTokenParser String InnerState InnerLowLevelMonad
innerLexer = Text.Parsec.Token.makeTokenParser innerLangDef

outerLexer :: Text.Parsec.Token.GenTokenParser String OuterState OuterLowLevelMonad
outerLexer = Text.Parsec.Token.makeTokenParser outerLangDef

iIdentifier    = Text.Parsec.Token.identifier    innerLexer
iWhiteSpace    = Text.Parsec.Token.whiteSpace    innerLexer
iParens        = Text.Parsec.Token.parens        innerLexer
iReservedOp    = Text.Parsec.Token.reservedOp    innerLexer

oIdentifier    = Text.Parsec.Token.identifier    outerLexer
oWhiteSpace    = Text.Parsec.Token.whiteSpace    outerLexer
oReservedOp    = Text.Parsec.Token.reservedOp    outerLexer
oReserved      = Text.Parsec.Token.reserved      outerLexer
oDecimal       = Text.Parsec.Token.decimal       outerLexer
oStringLiteral = Text.Parsec.Token.stringLiteral outerLexer

myRunParser :: (MonadError String m) => ParsecT String st (Either String) ret -> st -> String -> String -> m ret
myRunParser parser st document lang = case runParserT parser st "-" document of {
  Left nonParserErr -> throwError nonParserErr;
  Right result -> case result of {
    Left parserErr -> throwError $ lang ++ show (sourceLine $ errorPos parserErr) ++ ":" ++ show (sourceColumn $ errorPos parserErr) ++ ": " ++ Text.Parsec.Error.showErrorMessages "or" "???" "expected" "found" "nothing" (Text.Parsec.Error.errorMessages parserErr);
    Right result -> return result;
  };
}

e :: (MonadError String m) => ParsecT String st m a -> ParsecT String st m a
e a = catchError a (\err -> do {
  p <- getPosition;
  i <- getInput;
  throwError $ show (sourceLine p) ++ ":" ++ show (sourceColumn p) ++ ": " ++ err ++ "\n\n" ++ take 100 i;
})

iTerm :: InnerParserMonad Statement
iTerm = do {
  st <- getState;
  a <- exprParse $ stSyntax st;
  return a;
}

iParser :: InnerParserMonad Statement
iParser = do {
  iWhiteSpace;
  result <- iTerm;
  eof;
  return result;
}

oTerm :: OuterParserMonad Statement
oTerm = do {
  st <- getState;
  str <- oIdentifier <|> oStringLiteral;
  result <- e $ myRunParser iParser st str ("term \"" ++ str ++ "\":");
  return result;
}

oSyntax :: Statement -> ExprTable String InnerState InnerLowLevelMonad Statement -> OuterParserMonad (ExprTable String InnerState InnerLowLevelMonad Statement)
oSyntax cnst syntax = iParens (do {
    s <- oStringLiteral;
    return $ exprAddNormal 1000 (\f -> do {
      iReservedOp s;
      return $ cnst;
    }) syntax;
  } <|> do {
    oReserved "prefix";
    s <- oStringLiteral;
    priority <- oDecimal;
    return $ exprAddNormal (fromInteger priority) (\f -> do {
      iReservedOp s;
      arg <- f (fromInteger priority);
      e $ application cnst arg;
    }) syntax;
  } <|> do {
    oReserved "infix";
    s <- oStringLiteral;
    priority <- oDecimal;
    return $ exprAddNormal (fromInteger priority) (\f -> do {
      a <- f (fromInteger priority + 1);
      iReservedOp s;
      b <- f (fromInteger priority + 1);
      cnsta <- e $ application cnst a;
      e $ application cnsta b;
    }) syntax;
  } <|> do {
    oReserved "infixl";
    s <- oStringLiteral;
    priority <- oDecimal;
    return $ exprAddRecursive (fromInteger priority) (\f a -> do {
      iReservedOp s;
      b <- f (fromInteger priority + 1);
      cnsta <- e $ application cnst a;
      e $ application cnsta b;
    }) syntax;
  } <|> do {
    oReserved "infixr";
    s <- oStringLiteral;
    priority <- oDecimal;
    return $ exprAddNormal (fromInteger priority) (\f -> do {
      a <- f (fromInteger priority + 1);
      iReservedOp s;
      b <- f (fromInteger priority);
      cnsta <- e $ application cnst a;
      e $ application cnsta b;
    }) syntax;
  } <|> do {
    oReserved "binder";
    s <- oStringLiteral;
    priority <- oDecimal;
    return $ exprAddNormal (fromInteger priority) (\f -> do {
      st <- getState;
      iReservedOp s;
      x <- iIdentifier;
      iReservedOp "::";
      a <- f $ fromInteger priority;
      iReservedOp ".";
      xx <- e $ start (stN st) x a;
      setState $ st { stN = stN st + 1, stC = stC st ++ [(x, xx)] };
      b <- f $ fromInteger priority;
      setState st;
      lam <- e $ abstraction xx b;
      e $ application cnst lam;
    }) syntax;
  }) <|> return syntax

oStep :: OuterParserMonad [Statement]
oStep = do {
    st <- getState;
    oReserved "fix";
    i <- try (do {
      i <- oIdentifier;
      oReservedOp "::";
      return i;
    }) <|> return "this";
    a <- oTerm;
    a2 <- e $ start (stN st) i a;
    newSyntax <- oSyntax a2 (stSyntax st);
    setState $ st { stN = stN st + 1, stC = stC st ++ [(i, a2), ("this", a2), ("assm", a2)], stSyntax = newSyntax };
    return [a2];
} <|> do {
  st <- getState;
  oReserved "abbr";
  i <- try (do {
    i <- oIdentifier;
    oReservedOp "=";
    return i;
  }) <|> return "this";
  a <- oTerm;
  newSyntax <- oSyntax a (stSyntax st);
  setState $ st { stC = stC st ++ [(i, a), ("this", a)], stSyntax = newSyntax };
  return [];
} <|> do {
  stBegin <- getState;
  oReservedOp "{";
  llist <- many1 oStep;
  list <- Control.Monad.foldM (\x y -> return $ x ++ y) [] llist;
  oReservedOp "}";
  stEnd <- getState;
  a <- e $ runMaybe $ lookup "this" $ reverse $ stC stEnd;
  a2 <- e $ Control.Monad.foldM (flip abstraction) a (reverse $ list);
  setState $ stBegin { stC = stC stBegin ++ [("this", a2)] };
  return [];
}

oParser :: OuterParserMonad ()
oParser = do {
  oWhiteSpace;
  oReserved "sorts";
  ss <- many oIdentifier;
  oReserved "axioms";
  as <- many (do {
    a <- oIdentifier;
    oReservedOp "::";
    b <- oIdentifier;
    return (a, b);
  });
  oReserved "rules";
  rs <- many $ do {
    oReservedOp "(";
    a <- oIdentifier;
    oReservedOp ",";
    b <- oIdentifier;
    oReservedOp ",";
    c <- oIdentifier;
    oReservedOp ")";
    return (a, b, c);
  };
  c <- e $ createPTS ss as rs;
  setState $ InnerState { stN = 0, stC = c, stSyntax =
      exprAddNormal 1000 (\f -> do {
        st <- getState;
        i <- iIdentifier;
        case lookup i $ reverse $ stC st of {
          Nothing -> throwError $ i ++ ": undeclared";
          Just result -> return result;
        };
      }) $
      exprAddNormal 1000 (\f -> iParens (f 0)) $
      exprAddNormal 3 (\f -> do {
        a <- f 4;
        iReservedOp "::";
        t <- f 0;
        e $ typed a t;
        return a;
      }) $
      exprAddRecursive 999 (\f func -> do {
        arg <- f 1000;
        e $ application func arg;
      }) $
      exprAddNormal 3 (\f -> do {
        st <- getState;
        iReservedOp "%";
        x <- iIdentifier;
        iReservedOp "::";
        a <- f 0;
        iReservedOp ".";
        xx <- e $ start (stN st) x a;
        setState $ st { stN = stN st + 1, stC = stC st ++ [(x, xx)] };
        b <- f 3;
        setState st;
        e $ abstraction xx b;
      }) $
      exprAddNormal 0 (\f -> do {
        st <- getState;
        iReservedOp "!!";
        x <- iIdentifier;
        iReservedOp "::";
        a <- f 0;
        iReservedOp ".";
        xx <- e $ start (stN st) x a;
        setState $ st { stN = stN st + 1, stC = stC st ++ [(x, xx)] };
        b <- f 0;
        setState st;
        e $ arrow xx b;
      }) $
      exprAddNormal 1 (\f -> do {
        st <- getState;
        a <- f 2;
        iReservedOp "==>";
        b <- f 1;
        aa <- e $ start (stN st) "_" a;
        e $ arrow aa b;
      }) exprEmpty };
  many (do {
    oStep;
    return ();
  });
  eof;
  return ();
}

main = do {
  document <- getContents;
  case myRunParser oParser (InnerState { stN = 0, stC = [], stSyntax = exprEmpty }) document "" of {
    Left err -> error err;
    Right () -> return ();
  };
}
